// © 2023-2024 John Breaux
//! A [Token] is a [semantically-tagged](TokenKind) [sequence of characters](str) and a [Span]
//!
//! [Tokens](Token) are a borrowed, and cannot outlive their source slice (lifetime `'t`)
use crate::span::Span;
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct Token<'t> {
    pub lexeme: &'t str,
    pub kind: TokenKind,
    pub pos: Span<usize>,
}
impl<'t> Token<'t> {
    pub fn new(lexeme: &'t str, kind: TokenKind, pos: Span<usize>) -> Self {
        Self { lexeme, kind, pos }
    }
    pub fn kind(&self) -> TokenKind {
        self.kind
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum TokenKind {
    Eof,
    Newline,    // \n
    OpenParen,  // (
    CloseParen, // )
    OpenCurly,  // {
    CloseCurly, // }
    OpenBrace,  // [
    CloseBrace, // ]
    Comma,      // ,
    Colon,      // :
    Bang,       // !
    At,         // @
    Amp,        // &
    Bar,        // |
    Caret,      // ^
    Star,       // *
    Hash,       // #
    Dollar,     // $
    Percent,    // %
    Plus,       // +
    Minus,      // -
    Slash,      // /
    Lsh,        // <<
    Rsh,        // >>

    Comment,         // (';' | '//') .* '\n' |
    Directive,       // '.' XID_CONTINUE*
    Identifier,      // XID_START XID_CONTINUE*
    Number(u16, u8), // varies depending on base
    Char(char),      // '\'' ('\' Escape | .) '\''
    String,          // '"' .* '"'
    Reg(Reg),
    NoEm(NoEm),
    OneEm(OneEm),
    Special(Special),
    OneArg(OneArg),
    TwoArg(TwoArg),
    Jump(Jump),

    Byte, // .b
    Word, // .w
}
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Reg {
    PC,
    SP,
    SR,
    CG,
    R4,
    R5,
    R6,
    R7,
    R8,
    R9,
    R10,
    R11,
    R12,
    R13,
    R14,
    R15,
}
/// Fake instructions of the form `opcode`
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum NoEm {
    Nop,
    Ret,
    Clrc,
    Clrz,
    Clrn,
    Setc,
    Setz,
    Setn,
    Dint,
    Eint,
}
/// Fake instructions of the form `opcode dst`
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum OneEm {
    Pop,
    Rla,
    Rlc,
    Inv,
    Clr,
    Tst,
    Dec,
    Decd,
    Inc,
    Incd,
    Adc,
    Dadc,
    Sbc,
}
/// These opcodes have bespoke grammatical rules
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Special {
    /// Br = "br" Src
    Br,
}
/// Real instructions of the form `opcode src`
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum OneArg {
    Rrc,
    Swpb,
    Rra,
    Sxt,
    Push,
    Call,
    Reti,
}
/// Real instructions of the form `opcode src, dst`
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum TwoArg {
    Mov,
    Add,
    Addc,
    Subc,
    Sub,
    Cmp,
    Dadd,
    Bit,
    Bic,
    Bis,
    Xor,
    And,
}
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Jump {
    Jne,
    Jnz,
    Jeq,
    Jz,
    Jnc,
    Jlo,
    Jc,
    Jhs,
    Jn,
    Jge,
    Jl,
    Jmp,
}
mod convert {
    //! Implementations of [FromStr] for [token](super) types.
    use super::*;
    use std::str::FromStr;

    impl FromStr for Reg {
        type Err = ();
        fn from_str(s: &str) -> Result<Self, Self::Err> {
            Ok(match s {
                "pc" => Reg::PC,
                "sp" => Reg::SP,
                "sr" => Reg::SR,
                "cg" => Reg::CG,
                "r0" => Reg::PC,
                "r1" => Reg::SP,
                "r2" => Reg::SR,
                "r3" => Reg::CG,
                "r4" => Reg::R4,
                "r5" => Reg::R5,
                "r6" => Reg::R6,
                "r7" => Reg::R7,
                "r8" => Reg::R8,
                "r9" => Reg::R9,
                "r10" => Reg::R10,
                "r11" => Reg::R11,
                "r12" => Reg::R12,
                "r13" => Reg::R13,
                "r14" => Reg::R14,
                "r15" => Reg::R15,
                _ => Err(())?,
            })
        }
    }
    impl FromStr for NoEm {
        type Err = ();
        fn from_str(s: &str) -> Result<Self, Self::Err> {
            Ok(match s {
                "nop" => NoEm::Nop,
                "ret" => NoEm::Ret,
                "clrc" => NoEm::Clrc,
                "clrz" => NoEm::Clrz,
                "clrn" => NoEm::Clrn,
                "setc" => NoEm::Setc,
                "setz" => NoEm::Setz,
                "setn" => NoEm::Setn,
                "dint" => NoEm::Dint,
                "eint" => NoEm::Eint,
                _ => Err(())?,
            })
        }
    }
    impl FromStr for OneEm {
        type Err = ();
        fn from_str(s: &str) -> Result<Self, Self::Err> {
            Ok(match s {
                "pop" => OneEm::Pop,
                "rla" => OneEm::Rla,
                "rlc" => OneEm::Rlc,
                "inv" => OneEm::Inv,
                "clr" => OneEm::Clr,
                "tst" => OneEm::Tst,
                "dec" => OneEm::Dec,
                "decd" => OneEm::Decd,
                "inc" => OneEm::Inc,
                "incd" => OneEm::Incd,
                "adc" => OneEm::Adc,
                "dadc" => OneEm::Dadc,
                "sbc" => OneEm::Sbc,
                _ => Err(())?,
            })
        }
    }
    impl FromStr for Special {
        type Err = ();
        fn from_str(s: &str) -> Result<Self, Self::Err> {
            Ok(match s {
                "br" => Special::Br,
                _ => Err(())?,
            })
        }
    }
    impl FromStr for OneArg {
        type Err = ();
        fn from_str(s: &str) -> Result<Self, Self::Err> {
            Ok(match s {
                "rrc" => OneArg::Rrc,
                "swpb" => OneArg::Swpb,
                "rra" => OneArg::Rra,
                "sxt" => OneArg::Sxt,
                "push" => OneArg::Push,
                "call" => OneArg::Call,
                "reti" => OneArg::Reti,
                _ => Err(())?,
            })
        }
    }
    impl FromStr for TwoArg {
        type Err = ();
        fn from_str(s: &str) -> Result<Self, Self::Err> {
            Ok(match s {
                "mov" => TwoArg::Mov,
                "add" => TwoArg::Add,
                "addc" => TwoArg::Addc,
                "subc" => TwoArg::Subc,
                "sub" => TwoArg::Sub,
                "cmp" => TwoArg::Cmp,
                "dadd" => TwoArg::Dadd,
                "bit" => TwoArg::Bit,
                "bic" => TwoArg::Bic,
                "bis" => TwoArg::Bis,
                "xor" => TwoArg::Xor,
                "and" => TwoArg::And,
                _ => Err(())?,
            })
        }
    }
    impl FromStr for Jump {
        type Err = ();
        fn from_str(s: &str) -> Result<Self, Self::Err> {
            Ok(match s {
                "jne" => Jump::Jne,
                "jnz" => Jump::Jnz,
                "jeq" => Jump::Jeq,
                "jz" => Jump::Jz,
                "jnc" => Jump::Jnc,
                "jlo" => Jump::Jlo,
                "jc" => Jump::Jc,
                "jhs" => Jump::Jhs,
                "jn" => Jump::Jn,
                "jge" => Jump::Jge,
                "jl" => Jump::Jl,
                "jmp" => Jump::Jmp,
                _ => Err(())?,
            })
        }
    }
}
mod display {
    //! Implementations of [Display] for [token](super) types.
    use super::*;
    use std::fmt::Display;
    impl<'t> Display for Token<'t> {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            let Self { lexeme, kind, pos: _ } = self;
            match kind {
                TokenKind::Comment
                | TokenKind::Directive
                | TokenKind::Identifier
                | TokenKind::String => {
                    write!(f, "{}", lexeme)
                }
                ty => ty.fmt(f),
            }
        }
    }
    impl Display for TokenKind {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                TokenKind::Eof => write!(f, "[EOF]"),
                TokenKind::Newline => writeln!(f),
                TokenKind::OpenParen => write!(f, "("),
                TokenKind::CloseParen => write!(f, ")"),
                TokenKind::OpenCurly => write!(f, "{{"),
                TokenKind::CloseCurly => write!(f, "}}"),
                TokenKind::OpenBrace => write!(f, "["),
                TokenKind::CloseBrace => write!(f, "]"),
                TokenKind::Comma => write!(f, ","),
                TokenKind::Colon => write!(f, ":"),
                TokenKind::Bang => write!(f, "!"),
                TokenKind::At => write!(f, "@"),
                TokenKind::Amp => write!(f, "&"),
                TokenKind::Bar => write!(f, "|"),
                TokenKind::Caret => write!(f, "^"),
                TokenKind::Star => write!(f, "*"),
                TokenKind::Hash => write!(f, "#"),
                TokenKind::Dollar => write!(f, "$"),
                TokenKind::Percent => write!(f, "%"),
                TokenKind::Plus => write!(f, "+"),
                TokenKind::Minus => write!(f, "-"),
                TokenKind::Slash => write!(f, "/"),
                TokenKind::Lsh => write!(f, "<<"),
                TokenKind::Rsh => write!(f, ">>"),
                TokenKind::Comment => write!(f, "; "),
                TokenKind::Directive => write!(f, "."),
                TokenKind::Identifier => write!(f, "Identifier"),
                TokenKind::Number(val, 2) => write!(f, "0b{val:b}"),
                TokenKind::Number(val, 8) => write!(f, "0o{val:o}"),
                TokenKind::Number(val, 16) => write!(f, "0x{val:x}"),
                TokenKind::Number(val, _) => write!(f, "{val}"),
                TokenKind::Char(c) => write!(f, "'{c}'"),
                TokenKind::String => write!(f, "\"String\""),
                TokenKind::Reg(kw) => write!(f, "{kw}"),
                TokenKind::NoEm(kw) => write!(f, "{kw}"),
                TokenKind::OneEm(kw) => write!(f, "{kw}"),
                TokenKind::Special(kw) => write!(f, "{kw}"),
                TokenKind::OneArg(kw) => write!(f, "{kw}"),
                TokenKind::TwoArg(kw) => write!(f, "{kw}"),
                TokenKind::Jump(kw) => write!(f, "{kw}"),
                TokenKind::Byte => write!(f, ".b"),
                TokenKind::Word => write!(f, ".w"),
            }
        }
    }
    impl Display for Reg {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                Reg::PC => "pc".fmt(f),
                Reg::SP => "sp".fmt(f),
                Reg::SR => "sr".fmt(f),
                Reg::CG => "cg".fmt(f),
                Reg::R4 => "r4".fmt(f),
                Reg::R5 => "r5".fmt(f),
                Reg::R6 => "r6".fmt(f),
                Reg::R7 => "r7".fmt(f),
                Reg::R8 => "r8".fmt(f),
                Reg::R9 => "r9".fmt(f),
                Reg::R10 => "r10".fmt(f),
                Reg::R11 => "r11".fmt(f),
                Reg::R12 => "r12".fmt(f),
                Reg::R13 => "r13".fmt(f),
                Reg::R14 => "r14".fmt(f),
                Reg::R15 => "r15".fmt(f),
            }
        }
    }
    impl Display for NoEm {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                NoEm::Nop => "nop".fmt(f),
                NoEm::Ret => "ret".fmt(f),
                NoEm::Clrc => "clrc".fmt(f),
                NoEm::Clrz => "clrz".fmt(f),
                NoEm::Clrn => "clrn".fmt(f),
                NoEm::Setc => "setc".fmt(f),
                NoEm::Setz => "setz".fmt(f),
                NoEm::Setn => "setn".fmt(f),
                NoEm::Dint => "dint".fmt(f),
                NoEm::Eint => "eint".fmt(f),
            }
        }
    }
    impl Display for OneEm {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                OneEm::Pop => "pop".fmt(f),
                OneEm::Rla => "rla".fmt(f),
                OneEm::Rlc => "rlc".fmt(f),
                OneEm::Inv => "inv".fmt(f),
                OneEm::Clr => "clr".fmt(f),
                OneEm::Tst => "tst".fmt(f),
                OneEm::Dec => "dec".fmt(f),
                OneEm::Decd => "decd".fmt(f),
                OneEm::Inc => "inc".fmt(f),
                OneEm::Incd => "incd".fmt(f),
                OneEm::Adc => "adc".fmt(f),
                OneEm::Dadc => "dadc".fmt(f),
                OneEm::Sbc => "sbc".fmt(f),
            }
        }
    }
    impl Display for Special {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                Special::Br => "br".fmt(f),
            }
        }
    }
    impl Display for OneArg {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                OneArg::Rrc => "rrc".fmt(f),
                OneArg::Swpb => "swpb".fmt(f),
                OneArg::Rra => "rra".fmt(f),
                OneArg::Sxt => "sxt".fmt(f),
                OneArg::Push => "push".fmt(f),
                OneArg::Call => "call".fmt(f),
                OneArg::Reti => "reti".fmt(f),
            }
        }
    }
    impl Display for TwoArg {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                TwoArg::Mov => "mov".fmt(f),
                TwoArg::Add => "add".fmt(f),
                TwoArg::Addc => "addc".fmt(f),
                TwoArg::Subc => "subc".fmt(f),
                TwoArg::Sub => "sub".fmt(f),
                TwoArg::Cmp => "cmp".fmt(f),
                TwoArg::Dadd => "dadd".fmt(f),
                TwoArg::Bit => "bit".fmt(f),
                TwoArg::Bic => "bic".fmt(f),
                TwoArg::Bis => "bis".fmt(f),
                TwoArg::Xor => "xor".fmt(f),
                TwoArg::And => "and".fmt(f),
            }
        }
    }
    impl Display for Jump {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                Jump::Jne => "jne".fmt(f),
                Jump::Jnz => "jnz".fmt(f),
                Jump::Jeq => "jeq".fmt(f),
                Jump::Jz => "jz".fmt(f),
                Jump::Jnc => "jnc".fmt(f),
                Jump::Jlo => "jlo".fmt(f),
                Jump::Jc => "jc".fmt(f),
                Jump::Jhs => "jhs".fmt(f),
                Jump::Jn => "jn".fmt(f),
                Jump::Jge => "jge".fmt(f),
                Jump::Jl => "jl".fmt(f),
                Jump::Jmp => "jmp".fmt(f),
            }
        }
    }
}
